#include "screen.hpp"
#include "vec.hpp"
#include <cmath>
#include <cstdio>
#include <swarm.hpp>

static constexpr int   kFPS = 60;
static constexpr float kDT  = 1.0 / kFPS;

float f(vec<2> x) {
    return  50 * (std::pow(std::sin((x.x - 10)/2),2) + std::pow(std::sin(x.y/2),2)) +
	std::pow(std::abs(x.x - 3.141592653589), 1.2) + std::pow(std::abs(x.y), 1.2);
}

int
main(int argc, char **argv) {
    int iter_count;
    if(argc < 2) iter_count = 100;
    else sscanf(argv[1], "%d", &iter_count);

    Swarm<2> swarm(f, iter_count);
    Screen scr(80, 24);

    TanhColorizer tanh_col;

    float min_f = 0, max_f = 0;
    scr.shader = [&](int w, int h, float x, float y) -> Screen::Symbol {
	float z = f({x, y});
	min_f = std::min(min_f, z);
	max_f = std::max(max_f, z);
	return Screen::Symbol {
	    .sym = ' ',
	    .color = tanh_col(z),
	};
    };

    swarm.add_particle({0, 0}, {5, 3});
    swarm.add_particle({4, 1}, {1, 5});
    swarm.add_particle({16, 10}, {4, 5});
    swarm.add_particle({7, 4}, {4, -3});
    swarm.add_particle({14, 1}, {1, 5});
    swarm.add_particle({1, 10}, {9, 10});
    swarm.add_particle({5, 4}, {5, -3});
    swarm.add_particle({4, 6}, {-1, 5});
    swarm.add_particle({2, 6}, {7, 0});
    swarm.add_particle({7, 7}, {4, -3});

    for(const auto &p : swarm.get_particles()) {
	scr.points.push_back(&p.get_position());
    }

    enter_noncanonical_mode();
    // swap to alt buffer
    printf("\e[?1049h");

    bool pause = false;
    bool frame_step = false;

    // initialize colorizer scale/translate
    scr.draw();
    tanh_col.scale = std::max(std::abs(max_f), std::abs(min_f));
    tanh_col.translate = (max_f - min_f) / 2.0;

    auto update_and_draw = [&]() -> void {
	scr.draw();
    };

    // We draw to the screen at a rate of `kFPS', but step()ing the swarm at
    // this rate would be far too fast to be interesting to look at. On the
    // other hand, step()ing once a second is too slow.
    //
    // So, we step() every quarter second. To line up the move() so that a unit
    // move() happens preceding each step(), we also need to move as fast as
    // we've shortened our step() interval.
    for(int i = 0; i < iter_count * kFPS/4;) {
	if(!pause) {
	    scr.clear();

	    update_and_draw();

	    // Since we step() 4x a second, we need to move() 4x as fast as moving by dt
	    // each frame.
	    swarm.move(kDT * 4);
	    if(i % (kFPS/4) == (kFPS/4)-1)
		swarm.step();

	    ++i;
	}

	if(frame_step) {
	    pause = true;
	    frame_step = false;
	}

	struct {char buf[8]; int size;} inbuf = {{}, 0};
	inbuf.size = read(STDIN_FILENO, inbuf.buf, 7);
	switch (*inbuf.buf) {
	case 'q':
	    goto cleanup;
	case ' ':
	    pause = !pause;
	    break;
	case '.':
	    pause = false;
	    frame_step = true;
	    break;

	case 'd':
	    scr.dx += scr.sy;
	    update_and_draw();
	    break;
	case 'a':
	    scr.dx -= scr.sy;
	    update_and_draw();
	    break;
	case 's':
	    scr.dy += scr.sx;
	    update_and_draw();
	    break;
	case 'w':
	    scr.dy -= scr.sx;
	    update_and_draw();
	    break;

	case 'I':
	    tanh_col.scale *= 1.2;
	    update_and_draw();
	    break;
	case 'O':
	    tanh_col.scale /= 1.2;
            update_and_draw();
            break;

	case 'K':
	    tanh_col.translate *= 1.2;
	    update_and_draw();
	    break;
	case 'L':
	    tanh_col.translate /= 1.2;
            update_and_draw();
            break;

	case 'i':
	    scr.dx += 0.5 * (1 - 0.75) * scr.sx * scr.w;
	    scr.dy += 0.5 * (1 - 0.75) * scr.sy * scr.h;
	    scr.sx *= 0.75;
	    scr.sy *= 0.75;
	    update_and_draw();
	    break;
	case 'o':
	    scr.dx -= 0.5 * (1 / 0.75 - 1) * scr.sx * scr.w;
	    scr.dy -= 0.5 * (1 / 0.75 - 1) * scr.sy * scr.h;
	    scr.sx /= 0.75;
	    scr.sy /= 0.75;
	    update_and_draw();
	    break;

	case 'p':
	    auto [y, x] = swarm.best();
	    printf("Current best: f(%.10f, %.10f) = %.10f", x.x, x.y, y);
	    break;
	}

	// Wait for kDT (time between two frames)
	usleep(1000 * 1000 / kFPS);
    }

cleanup:
    // swap back to original buffer
    printf("\e[?1049l");
    enter_canonical_mode();

    auto [y, x] = swarm.best();
    printf("Best: f(" V2_FMT ") = %.3f", V2_ARG(x), y);

    return 0;
}
